/**/

   /***************************************************************************************************/
   /*  rpmDBlist.cmd                                                                                  */
   /*  Anton Monroe <amonroe5301@gmail.com>                                                           */
   /*                                                                                                 */
   /*  extracts a list of all files and their packages from the rpm database                          */
   /*  (by calling rpm.exe) and writes it to a file. Normally '/var/lib/rpm.list'.                    */
   /*                                                                                                 */
   /*  The list is used by 'whatowns.btm' because it is much faster than calling rpm.exe              */
   /*  The format of each line of rpm.list is <filename><separator><package name>                     */
   /*                                                                                                 */
   /*  This assumes it is running under 4OS2. (it uses the *ren command)                              */
   /*                                                                                                 */
   /*  I normally run this as a Cron job every day. If the rpm database has not                       */
   /*  changed, this does nothing but change the timestamp of rpm.list                                */
   /***************************************************************************************************/

   Main: /*fold00*/
   /*=================================================================================================*/
   /*=================================================================================================*/
   UnixRoot = OS2var('unixroot')
   /*  variables you may want to change:                                                              */
   /*  rpm.exe does not seem to have a way to query where the database is,                            */
   /*  so we assume the default: '/@unixroot/var/lib/rpm/Packages'                                    */
   /*  note that the text version is in /var/lib/, not /var/lib/rpm/                                  */
   rpmDB = UnixRoot'\var\lib\rpm\Packages'
   rpmList = UnixRoot'\var\lib\rpm.list'

   G. = ''
   /* ----------------------------------------------------------------------------------------------- */
   parse source . . G.0me
   call SysFileTree G.0me, 'tmp.', 'FO'
   G.0me_long = tmp.1
   G.0me = filespec('name', G.0me_long)
   G.0meID = '['G.0me']'
   G.0me_ind = copies(' ', length(G.0me))
   G.0meID_ind = copies(' ', length(G.0meID))
   drop tmp.
   /* ----------------------------------------------------------------------------------------------- */

   /*  Sep is the character that divides filename and packagename. It must be the second              */
   /*  character of the header so WhatOwns.btm can find it.                                           */
   /*  I cannot use Tab because 4OS2 eats it, somehow. I don't want to use a control character        */
   /*  because it confuses file viewers/editors. I could use any character that is not                */
   /*  legal in a filename, but an upper-ascii graphics character is more visible if I                */
   /*  want to look at the file for some reason.                                                      */
   G.0Sep = 'B2'x
   OutHeader = '#'||G.0Sep||' made by 'G.0me' from 'rpmDB' at 'DateStr()' 'time()
   /*                                                                                                 */
   /*  the default log file if not given on commandline                                               */
   Logfile = 'STDERR'
   /* =============================================================================================== */
   /* Load RexxUtil Library                                                                           */
   if RxFuncQuery('SysLoadFuncs') then do
         call RxFuncAdd 'SysLoadFuncs', 'RexxUtil', 'SysLoadFuncs'
         call SysLoadFuncs
      end

   call on Error
   call on Failure
   call on NotReady
   signal on NoValue
   /***************************************************************************************************/

   /*    variables that are likely to be used by many procedures                                      */
   Shared_vars = 'SIGL Opt. G. logfile rpmList Unixroot Outheader FileList.'

   Opt. = ''
   parse arg args
   interpret GetOpts(args)

   if Opt.0help \= '' then do
         call Help
         call Egress 'usage'
      end

   if UnixRoot == '' then do
         call LogIt '[ERROR] OS/2 variable UnixRoot not found'
         call Egress 'usage'
      end

   if Opt.0logfile \= '' then do
         testpath = filespec('path', Opt.0logfile)
         if  \DirExist(testpath) then do
               call lineout 'stderr', G.0meID '[Error] logfile is' Opt.0logfile', but directory' ,
                    testpath 'does not exist'
               call Egress 'usage'
            end
         Logfile = Opt.0logfile
      end
   /*  from here on, error messages go to logfile                                                     */

   if  \FileExist(rpmDB) then do
         call LogIt '[Error] cannot find rpm database at 'rpmDB
         call Egress 'usage'
      end

   /*    For debugging, I can override the rpmList on command line                                    */
   if Opt.0file \= '' then rpmList = Opt.0file

   outpath = filespec('drive', rpmList)||filespec('path', rpmList)
   if \DirExist(outpath) then do
         call LogIt '[Error] cannot find directory "'outpath'" for output file'
         call Egress 'usage'
      end

   rpmDBTime = SysGetFileDateTime(rpmDB)
   rpmListTime = SysGetFileDateTime(rpmList)

   if rpmListTime > rpmDBTime then do
         /*  rpmDB has not changed since list was built--                                             */
         /*  update timestamp to make it clear that the list is up to date                            */
         call SysSetFileDateTime rpmList
         call LogIt 'No change; updated timestamp of 'rpmList
         call Egress 'okay'
      end

   call LogIt 'rpmList will be 'rpmList

   /*  call rpm.exe to put list into stem variable FileList.                                          */
   /*  change '@unixroot' to real path                                                                */
   /*  then sort list                                                                                 */
   ret = RpmToArray()
   if Opt.0debug > 0 then call LogIt '[Debug] RpmToArray() returned "'ret'"'
   if ret \= 0 then call Egress 'error'

   /*    back up old list                                                                             */
   rpmListbak = rpmList||'.bak'
   if FileExist(rpmList) then do
         call SysFileDelete rpmListbak
         '*ren /q 'rpmList rpmListbak
      end

   ret = ArrayToRPMList(rpmList, 'FileList.')
   if Opt.0debug > 0 then call LogIt '[Debug] ArrayToRPMList() returned "'ret'"'
   if ret == 9 then call Egress 'bug'
   if ret \= 0 then call Egress 'error'
   call LogIt 'made new list in 'rpmList
   call Egress 'okay'
   /* === End of Main =============================================================================== */

   ArrayToRPMList: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  writes the FileList. array to file                                                             */
   /*  In: <file name>, <variable name>                                                               */
   /*  the stem =name= is passed, but the variable must be shared                                     */
   /* =============================================================================================== */
   parse arg outfile , stemN

   if stemN == '' then return 9
   listN = value('stemN')
   if listN == '' then return 9
   ListHigh = value(listN||0)

   /*    write the new file                                                                           */
   call lineout outfile, OutHeader
   do f = 1 to ListHigh
      thisline = value(listN||f)
      call lineout outfile, thisline
   end
   call lineout outfile
   return 0
   /* === End of ArrayToRPMList ===================================================================== */

   RpmToArray: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  get list of files and owners from rpm.exe; un-unixroot their paths, alphabetize list           */
   /*  Returns: list in FileList array                                                                */
   /* =============================================================================================== */
   rpmArgs = '-q -v -a --queryformat'
   /*    rpmArgs = '-q -a --queryformat'                                                              */
   QString = QfOwners()

   /*  make sure session queue is empty                                                               */
   do while queued() > 0
      dummy = linein('queue:')
   end

   QString = '"'QString'"'
   if Opt.0debug > 0 then call LogIt '[Debug]  will do "rpm.exe 'rpmArgs QString'| rxqueue.exe /fifo"'
   'rpm.exe 'rpmArgs QString' | rxqueue.exe /fifo '
   if Opt.0debug > 0 then call LogIt '[Debug]  "rpm.exe 'rpmArgs QString'| rxqueue.exe /fifo" done'

   atUnix = '/@unixroot'
   atLen = length(atUnix)

   /*  first read lines from queue into array                                                         */
   FileList. = ''
   FileList.0 = 0
   f = 0
   do while queued() > 0
      f = f +1
      FileList.f = linein('queue:')
   end
   FileList.0 = f

   if Opt.0debug > 0 then call LogIt '[Debug] FileList.0 == 'FileList.0
   if FileList.0 < 5 then do
         call LogIt '[Error] FileList.0 == 'FileList.0
         return 9
      end

   /*  then change rpm's "/@unixroot" string to the value of unixroot                                 */
   do f = 1 to FileList.0
      select
         when left(FileList.f, atLen) == atUnix then
            FileList.f = UnixRoot||substr(FileList.f, atLen +1)
         when left(FileList.f, 1) == '/' then
            /*  I found at least one entry without @unixroot, ie "/etc/samba/private"                 */
            FileList.f = UnixRoot||FileList.f
         otherwise do
               /*    show error message and quit                                                      */
               call LogIt '[ERROR] invalid line from rpm.exe: "'FileList.f'"'
               return 9
            end
      end   /*  select  */
   end f

   /*    then sort list alphabetically                                                                */
   ret = 0
   ret = SysStemSort('FileList.', 'A', 'I')
   if ret \= 0 then do
         call LogIt '[ERROR] SysStemSort returned 'ret
         return 9
      end
   return 0
   /* === End of RpmToArray ========================================================================= */

   QfOwners: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  returns a format string Set.QF in the form <filename><separator><full package name>            */
   /*  for an explanation of how this works, see rpminfo.cmd                                          */
   /* =============================================================================================== */

   Set. = ''
   Set.prefix = '['
   Set.suffix = '\n]'
   tag. = ''

   /*  FILENAMES                                                                                      */
   tag.ID = '{FILENAMES}'
   tag.width = ''
   tag.prefix = ''
   tag.suffix = G.0Sep
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   /*  NEVRA is the package name                                                                      */
   tag.ID = '{=NEVRA}'
   tag.width = ''
   tag.prefix = ''
   tag.suffix = ''
   tag.qf = tag.prefix||'%%'||tag.width||tag.ID||tag.suffix
   Set.qf = Set.qf||tag.qf

   Set.qf = Set.prefix||Set.qf||Set.suffix
   return Set.qf
   /* === End of QfOwners =========================================================================== */

   LogIt: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*    if Opt.0logfile is defined, saves some progress/debugging messages to a log file             */
   /*    automatically adds timestamp and line number                                                 */
   /* =============================================================================================== */
   parse arg logstr
   if Opt.0debug > 0 then
      logstr = '['date() time()']['G.0me':'sigl']' logstr
   else
      logstr = '['date() time()']' G.0meID logstr
   call lineout Logfile, logstr
   call lineout Logfile
   return
   /* === End of LogIt ============================================================================== */

   Help: /*fold00*/
   /* =============================================================================================== */
   /* =============================================================================================== */
   /*    call lineout 'stderr', ''                                                                    */
   call lineout 'stderr', G.0me
   call lineout 'stderr', '  converts the rpm database ('rpmDB') to a text version'
   call lineout 'stderr', '  ('rpmList') usable by whatowns.btm'
   call lineout 'stderr', ''
   call lineout 'stderr', 'Usage:'
   call lineout 'stderr', '   'G.0me '[options]'
   call lineout 'stderr', 'The options are mostly for testing:'
   call lineout 'stderr', '    -file=<filename>'
   call lineout 'stderr', '            write file list to <filename>'
   call lineout 'stderr', '            default is 'rpmList
   call lineout 'stderr', '    -logfile=<filename>'
   call lineout 'stderr', '            send error and debug messages to <filename> instead of stderr'
   call lineout 'stderr', '            default is' Logfile
   call lineout 'stderr', '    -debug'
   call lineout 'stderr', '            show a few debugging messages'
   return
   /* === End of Help =============================================================================== */

   NoValue: /*fold00*/
   0Novalue:
   /* =============================================================================================== */
   /*  signal on NoValue                                                                              */
   /* ----------------------------------------------------------------------------------------------- */
   /*  depends on variables: G.0meID sigl                                                             */
   /*  depends on functions: Egress()                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   signal off NoValue
   _ = sigl
   call lineout 'stderr', G.0meID'[NOVALUE Handler] at line '_' for variable "'condition('d')'"'
   call lineout 'stderr', '   ==> {'sourceline(_)'}'
   call Egress 'bug_NoValue'
   /* === End of 0Novalue =========================================================================== */

   Error: /*fold00*/
   0Error:
   /* =============================================================================================== */
   /*  signal on error                                                                                */
   /* ----------------------------------------------------------------------------------------------- */
   /*  this does not do much, but it is not used much anyway                                          */
   /* =============================================================================================== */
   signal off Error
   call lineout 'stderr', '[ErrorHandler] 'rc' at line 'sigl
   call lineout 'stderr', '       Error # 'rc' is "'errortext(rc)'"'
   call lineout 'stderr', '       for more help, type "helpmsg REX'rc'"'
   call Egress 'error'
   /* === End of 0Error ============================================================================= */

   Failure: /*fold00*/
   /* =============================================================================================== */
   /* =============================================================================================== */
   Esigl = sigl
   parse arg errmsg
   call LogIt  '[Failure] 'rc' from line 'Esigl': "'errmsg'"'
   call Egress 'error'
   /* === End of Failure ============================================================================ */

   NotReady: /*fold00*/
   /* =============================================================================================== */
   /* =============================================================================================== */
   Esigl = sigl
   parse arg errmsg
   call LogIt  '[Notready] 'rc' from line 'Esigl': "'errmsg'"'
   call Egress 'bug'
   /* === End of NotReady =========================================================================== */

   LineNo: /*fold00*/
   0LineNo:
   /* =============================================================================================== */
   /*  returns the number of the current line--    say 'now at line 'LineNo()                         */
   /* =============================================================================================== */
   _ = sigl
   return _
   /* === End of 0LineNo ============================================================================ */

   Mark: /*fold00*/
   0Mark:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       CALL Mark [<list of variable names>] [, <string>]                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  Shows the filename and line number it was called from, the current values of the               */
   /*  variables named, and an informational string                                                   */
   /*  Arg(1) is a list of variable names                                                             */
   /*  Arg(2) can be any descriptive text                                                             */
   /*  Examples:                                                                                      */
   /*      CALL mark 'varA varB', 'At top of the "DO x = 1 to 5" loop'                                */
   /*  If no variable names are given, it will just show a message:                                   */
   /*      CALL mark '', 'At top of the "DO x = 1 to 5" loop'                                         */
   /*                                                                                                 */
   /*  This must share all the variables of whatever routine it was called from.                      */
   /*  I cannot think of a way to do it without setting at least one variable here,                   */
   /*  so reserve the variable '_' for the use of Mark and similar routines.                          */
   /* =============================================================================================== */
   _ = sigl
   call lineout 'stderr', ''
   if arg(2, 'e') ,
    & arg(2) = '' then nop
   else call lineout 'stderr', '[Mark]['value('G.0me')':'_']' arg(2)
   do _ = 1 to 99
      if word(arg(1), _) = '' then leave _
      if symbol(word(arg(1), _)) = 'VAR' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||'= "'||value(word(arg(1), _))||'"'
      if symbol(word(arg(1), _)) = 'LIT' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||' is not defined'
      if symbol(word(arg(1), _)) = 'BAD' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||' is not a variable'
   end _
   drop _
   return 0
   /* === End of 0Mark ============================================================================== */

   Egress: /*fold00*/
   0Egress:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     call Egress <exit type>                                                                     */
   /* ----------------------------------------------------------------------------------------------- */
   /*  All exits come through here.                                                                   */
   /*  If you need to, say, clean up temporary files, do it from here                                 */
   /*                                                                                                 */
   /*  OKAY    == normal exit                                                                         */
   /*  USAGE   == user error                                                                          */
   /*  ERROR   == something else wrong                                                                */
   /*  BUG     == bug in the program                                                                  */
   /* =============================================================================================== */
   signal off NoValue
   signal off Syntax
   parse upper arg ExitCode

   /*    handle my typing mistakes                                                                    */
   if ExitCode = 'OKAY' then exit 0
   if ExitCode = '' then exit 0
   if ExitCode = 0 then exit 0
   if abbrev(ExitCode, 'HALT') then exit 1
   if abbrev(ExitCode, 'USAGE') then exit 1
   if abbrev(ExitCode, 'ERROR') then exit 3
   if abbrev(ExitCode, 'BUG') then exit 9
   /*    in case I used some other exit string and forgot to add it here:                             */
   exit 99
   /* === End of 0Egress ============================================================================ */

   DirExist: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*      DirExist(<directory name>)                                                                 */
   /*                                                                                                 */
   /*  Examples:                                                                                      */
   /*      if DirExist('F:\os2') then ...                                                             */
   /*      if \DirExist('F:\NotThere') then ...                                                       */
   /*      if \DirExist(filespec('path', outfile)) then ...                                           */
   /*                                                                                                 */
   /*  returns 1 if a directory exists, 0 if it does not                                              */
   /*                                                                                                 */
   /*  use either \ or /                                                                              */
   /*  trailing slashes are stripped                                                                  */
   /*      'F:\OS2' or 'F:\OS2\'   both mean 'F:\OS2'                                                 */
   /*      'F:' or 'F:\'           both mean 'F:\'                                                    */
   /*                                                                                                 */
   /*  in 4os2.exe/cmd.exe,                                                                           */
   /*      'F:'                    means "the current directory on drive F:"                          */
   /*  but that would be pointless here; the current directory must exist                             */
   /*  likewise for                                                                                   */
   /*      "DirExist('.')"                                                                            */
   /*                                                                                                 */
   /*  if no directory name is provided at all, it is taken as '.'                                    */
   /*  that way a calling routine can verify the path of a filename without the                       */
   /*  extra step of checking the output of filespec() for ''.                                        */
   /*          if \DirExist(filespec('path', outfile)) then do                                        */
   /*              say 'invalid file name, no such directory'                                         */
   /*              return 1                                                                           */
   /*          end                                                                                    */
   /*                                                                                                 */
   /*  relative paths like "..\..\bin" with leading ".." are legal, but                               */
   /*  SysFileTree() can easily be confused by them                                                   */
   /*                                                                                                 */
   /*  wildcards are legal in <directory name>, but probably not useful                               */
   /* =============================================================================================== */

   /* ----------------------------------------------------------------------------------------------- */
   /*  the root directory of a drive needs special handling because SysFileTree()                     */
   /*  will not find it unless it has at least one subdirectory.                                      */
   /*                                                                                                 */
   /*  There is a logic to that--                                                                     */
   /*  to SysFileTree(), "F:\OS2" means "the directory named F:\OS2".                                 */
   /*  An ending slash, "F:\OS2\" means "all subdirectories of F:\OS2".                               */
   /*  so there is no way to refer to the root directory of a drive.                                  */
   /*  "R:\" means "all subdirectories under R:\", and                                                */
   /*  "R:" means "the current directory on R:".                                                      */
   /* ----------------------------------------------------------------------------------------------- */
   parse arg DirName .
   if DirName == '' then return 1
   if DirName == '.' then return 1
   DirName = translate(DirName, '\', '/')

   /*  if DirName = 'F:\' verify that drive F: exists. If so, it presumably has a root directory      */
   if right(DirName, 2) == ':\' then do
         DriveN = filespec('drive', DirName)
         if SysDriveInfo(DriveN) = '' then
            return 0
         else
            return 1
      end
   else do
         DirName = strip(DirName, 't', '\')
         ret = SysFileTree(DirName, 'Dirlist.', 'do')
         if Dirlist.0 > 0 then
            return 1
         else
            return 0
      end
   /* === End of DirExist =========================================================================== */

   FileExist: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        if FileExist(<filename>) then                                                            */
   /*                                                                                                 */
   /*  returns 1 (True) if <filename> exists and 0 (False) if it does not.                            */
   /*                                                                                                 */
   /*  <filename> must be a real file. This returns 0 for devices or for streams like STDOUT.         */
   /*  Granted, STDOUT does exist, but it is more useful this way. I mostly use this when I want      */
   /*  to avoid clobbering an existing file,                                                          */
   /*                                                                                                 */
   /*  The main reason for this is that                                                               */
   /*        if FileExist(filename) then                                                              */
   /*  is easier to remember than                                                                     */
   /*         if stream(filename, 'c', 'query exists') then                                           */
   /* =============================================================================================== */
   /*  Most of the complications are because OS/2 Rexx and Regina return different values             */
   /*  from stream() for stdout and for devices. Regina usually returns 0 for stdout, but             */
   /*  not always. I don't know why.                                                                  */
   parse arg FileN
   if FileN = '' then return 0
   if wordpos(translate(FileN), 'STDOUT STDERR') > 0 then return 0
   ret = stream(FileN, 'c', 'query exists')
   if ret == '' then return 0
   if left(translate(ret), 5) == '\DEV\' then return 0
   if left(translate(ret), 5) == '/DEV/' then return 0
   return 1
   /* === End of FileExist ========================================================================== */

   FileSize: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        size = FileSize(<filename>)                                                              */
   /*                                                                                                 */
   /*    returns the size of a file.                                                                  */
   /*    if the file does not exist, returns ''                                                       */
   /*                                                                                                 */
   /*    Because                                                                                      */
   /*        "if FileSize(filename)"                                                                  */
   /*    is easier to use than                                                                        */
   /*        "if stream(filename,'c','query size') "                                                  */
   /*                                                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*         note that if the file does not exist FileSize() returns '', so                          */
   /*                "if FileSize(filename) == 0 then"                                                */
   /*         will be false if filename does not exist                                                */
   /*         what is usually more useful is                                                          */
   /*                "if FileSize(filename) > 0 then"                                                 */
   /*         or                                                                                      */
   /*                "if FileSize(filename) < 1 then"                                                 */
   /*         or                                                                                      */
   /*                "if FileSize(filename) <= 0 then"                                                */
   /* =============================================================================================== */
   parse arg Fname
   if stream(Fname, 'c', 'query exists') == '' then
      return ''
   else
      return stream(Fname, 'c', 'query size')
   /* === End of FileSize =========================================================================== */

   DateStr: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = DateStr()                                                                       */
   /*    returns date formatted as "2019-07-01"                                                       */
   /* =============================================================================================== */
   DateStr = date('s')
   parse var DateStr yr +4 mo +2 day .
   DateStr = yr'-'mo'-'day
   return DateStr
   /* === End of DateStr ============================================================================ */

   GetOpts: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     INTERPRET GetOpts(args [,valid])                                                            */
   /*                                                                                                 */
   /*  where 'args' is the command line to be parsed, and                                             */
   /*  'valid' is an optional space-separated list of valid switches                                  */
   /*  'help' and '?' are always valid                                                                */
   /*                                                                                                 */
   /*  Example:                                                                                       */
   /*     parse arg args                                                                              */
   /*     interpret GetOpts(args, 'switchA switchB')                                                  */
   /*                                                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  Switches must start with '-'. For example:                                                     */
   /*                                                                                                 */
   /*         -Foo                (Opt.0Foo == 1)                                                     */
   /*         -Foo=1              (Opt.0Foo == 1)                                                     */
   /*         -!Foo or -Foo=0     (Opt.0Foo == 0)                                                     */
   /*         -Foo=string         (Opt.0Foo == 'string')                                              */
   /*         -Foo="string"       (Opt.0Foo == 'string')                                              */
   /*         -Foo=a,b,c          (Opt.0Foo == 'a,b,c')                                               */
   /*         -Foo="a, b, c"      (Opt.0Foo == 'a, b, c')                                             */
   /*         -Foo="a b c"        (Opt.0Foo == 'a b c')                                               */
   /*         -Foo=123            (Opt.0Foo == '123')                                                 */
   /*         -Foo= or -Foo=''    (Opt.0Foo == '')                                                    */
   /*                                                                                                 */
   /*  switches must not have spaces before or after the '='                                          */
   /*  values with spaces must be in quotes. the quotes will be stripped.                             */
   /*                                                                                                 */
   /*  a switch without a value is set to '1'                                                         */
   /*        -switchA                                                                                 */
   /*  is the same as                                                                                 */
   /*        -switchA=1                                                                               */
   /*                                                                                                 */
   /*  a switch can be negated with "!", so                                                           */
   /*        -!switchA                                                                                */
   /*  is the same as                                                                                 */
   /*        -switchA=0                                                                               */
   /*                                                                                                 */
   /*     -? is the equivalent of -help                                                               */
   /*     they will both result in:                                                                   */
   /*        Opt.0help    == 1                                                                        */
   /*                                                                                                 */
   /*  Interpreting what GetOpts returns will create several variables that the                       */
   /*  caller can use--                                                                               */
   /*     Opt.              is a stem variable that holds whatever was specified by switches          */
   /*     Opt.0optlist      a space-separated list of options that were read                          */
   /*     Opt.0errors       number of errors; that is, when GetOpts() did not understand something    */
   /*                                                                                                 */
   /*                       elements of the Opt. compound variable all start with '0', to avoid       */
   /*                       name conflicts with simple variables                                      */
   /*                                                                                                 */
   /*     ArgV.             is a stem that holds the non-switch arguments                             */
   /*                       (ArgV.0 is the name of the .cmd script, NOT the number of items           */
   /*                       in the stem)                                                              */
   /*     ArgC              is the number of non-switch arguments                                     */
   /*                                                                                                 */
   /*  Example:                                                                                       */
   /*     Foo.cmd -dir=up -msg="that way" -v "From here"                                              */
   /*     will result in these variables:                                                             */
   /*        Opt.0optlist == opt.0dir opt.0msg                                                        */
   /*        Opt.0dir     == up                                                                       */
   /*        Opt.0msg     == that way                                                                 */
   /*        Opt.0v       == 1                                                                        */
   /*        ArgV.0       == Foo.cmd                                                                  */
   /*        ArgV.1       == From here                                                                */
   /*        ArgC         == 1                                                                        */
   /*                                                                                                 */
   /*  in the case of errors, GetOpts() will set Opt.0errors to the number of errors                  */
   /*  found, and cause the caller to show a message about it on stderr. The caller                   */
   /*  should probably abort if Opt.0errors is greater than 0.                                        */
   /* =============================================================================================== */
   /* Some important variables (inside GetOpts) are:                                                  */
   /*    Str.SetOpts       == commands the the caller will execute to set options                     */
   /*                          in the Opt. stem variable                                              */
   /*    Str.SetArgs       == commands the the caller will execute to put the arguments               */
   /*                          in the ArgV. and ArgC variables                                        */
   /*    Str.Errors        == commands the caller will execute to report errors                       */
   /*                          reading the command line                                               */
   /*    ReturnStr         == all of the above, combined and returned to caller                       */
   /*    flag.             == variables for GetOpt's internal use                                     */
   /*                          flag.error == number of errors trying to read command line             */
   /*    Str.SwitchList    == list of options (switches) used on command line                         */
   /*    ArgC              == number of non-switch arguments                                          */
   /*    ArgV.             == array of non-switch arguments                                           */
   /*    valid             == list of valid switches                                                  */
   /*                                                                                                 */
   /* =============================================================================================== */
   flag. = ''
   flag.errors = 0
   str. = ''
   ReturnStr = ''
   ArgC = 0
   ArgV. = ''
   parse arg args , valid
   do i = 1 to WordsQ(args)
      thisArg = WordQ(args, i)
      if left(thisArg, 1) == '-' then
         call ReadSwitch thisArg
      else do
            /*    not a switch, add to Arg list                                                       */
            ArgC = ArgC +1
            ArgV.ArgC = thisArg
         end
   end
   /* ----------------------------------------------------------------------------------------------- */
   /*    now put it all together into a series of commands for the caller to interpret                */
   if flag.errors > 0 then do
         str.SwitchList = strip(str.SwitchList' Opt.0Errors')
      end
   str.SetOpts = str.SetOpts' Opt.0Errors = 'flag.errors ';'
   str.SetOpts = str.SetOpts 'Opt.0optlist = "'str.SwitchList'" ;'

   str.SetArgs = 'ArgV. = "" ; parse source . . ArgV.0 ;  ArgC = "'ArgC'" ;'
   do i = 1 to ArgC
      varVal = Unquoted(ArgV.i)
      str.SetArgs = str.SetArgs' ArgV.'i' = 'Quoted(varVal)' ;'
   end

   ReturnStr = str.errors str.SetOpts str.SetArgs
   return ReturnStr
   /* === End of GetOpts ============================================================================ */

   ReadSwitch: procedure expose flag. str. valid (Shared_vars)
   /* =============================================================================================== */
   /*  this is part of GetOpts()                                                                      */
   /* =============================================================================================== */
   parse arg thisArg
   Negate = 0
   switch = strip(thisArg, 'L', '-')
   if left(switch, 1) == '!' then do
         switch = strip(switch, 'L', '!')
         Negate = 1
      end
   parse var switch optName '=' optVal
   select
      /* -? and -help have fixed meanings and are allowed even if not included in 'valid'             */
      when translate(optName) == '?' then
         optName  = 'Opt.0Help'
      when translate(optName) == 'HELP' then
         optName  = 'Opt.0Help'
      when optName == '' then do
            str.errors = ,
              str.errors||'call lineout "stderr", "[GetOpts] `-` without a switch name" ;'
            flag.errors = flag.errors +1
            return
         end
      when valid \= '' ,
         & wordpos(translate(optName), translate(valid)) == 0 then do
            str.errors = str.errors|| ,
                         'call lineout "stderr", "[GetOpts] I do not understand switch `'thisArg'`" ;'
            flag.errors = flag.errors +1
            return
         end
      otherwise
         optName = 'Opt.0'optName
   end   /*  select  */

   select
      when Negate then
         if optVal == '' then
            optVal = 0
         else do
               str.errors = str.errors|| ,
                            'call lineout "stderr", "[GetOpts] I do not understand switch `'thisArg'`" ;'
               flag.errors = flag.errors +1
               return
            end
      when optVal == '' ,
         & right(switch, 1) \= '=' then
         /*  "-switch=" means Opt.0switch == '', not 1                                                */
         optVal = '1'
      otherwise
         /*    try to be tolerant of unusual arguments; only strip quotes if they are balanced        */
         /*    and only one level of nested quotes                                                    */
         optVal = Unquoted(optVal)
   end   /*  select  */

   if wordpos(optName, str.SetOpts) > 0 then do
         str.errors = ,
           str.errors||'call lineout "stderr", "[GetOpts] duplicate switch `'thisArg'`" ;'
         flag.errors = flag.errors +1
      end
   else do
         str.SetOpts = str.SetOpts' 'optName' = 'Quoted(optVal)' ;'
         str.SwitchList = strip(str.SwitchList' 'optName)
      end
   return
   /* === End of ReadSwitch ========================================================================= */

   OS2var: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        VarValue = OS2var(<variable name>)                                                       */
   /*        OS2Path = OS2var('PATH')                                                                 */
   /*                                                                                                 */
   /*     returns value of a variable in the OS/2 environment,                                        */
   /*     or '' if not defined                                                                        */
   /* =============================================================================================== */
   parse arg Pname
   return value(Pname, , 'OS2ENVIRONMENT')
   /* === End of OS2var ============================================================================= */

   Quoted: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = Quoted(string)                                                                  */
   /*                                                                                                 */
   /*    if possible, quotes a string with the kind of quotes it does not contain                     */
   /*        {He said "no"}   --> {'He said "no"'}                                                    */
   /*        {He said 'no'}   --> {"He said 'no'"}                                                    */
   /*    otherwise uses double quotes and lets the caller deal with the errors                        */
   /* =============================================================================================== */
   parse arg string
   select
      when pos('"', string) == 0 then
         newstring = '"'string'"'
      when pos("'", string) == 0 then
         newstring = "'"string"'"
      otherwise
         newstring = '"'string'"'
   end
   return newstring
   /* === End of Quoted ============================================================================= */

   Unquoted: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = Unquoted(string)                                                                */
   /*                                                                                                 */
   /*    strips quotes from a string if and only if it is enclosed in a balanced pair                 */
   /*    of single or double quotes                                                                   */
   /* =============================================================================================== */
   parse arg string
   if length(string) < 2 then return string
   select
      when left(string, 1) == '"' ,
         & right(string, 1) == '"' then
         string = substr(string, 2, length(string) -2)
      when left(string, 1) == "'" ,
         & right(string, 1) == "'" then
         string = substr(string, 2, length(string) -2)
      otherwise
         nop
   end
   return string
   /* === End of Unquoted =========================================================================== */

   WordQ: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     foo = WordQ(<string>, <word number>, <separator characters>, <quote characters>)            */
   /*                                                                                                 */
   /*  The last two parameters are optional, so it will work                                          */
   /*  as a replacement for the Rexx Word() function.                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  A better Word() function--                                                                     */
   /*  It recognizes a quoted string as one word                                                      */
   /*  Optional parameters let you decide which characters                                            */
   /*  count as word separators and as quotes.                                                        */
   /*                                                                                                 */
   /*  Default word separators are <space> and <tab>                                                  */
   /*  Default quote characters are single, double, and                                               */
   /*  back quotes-- ' " `                                                                            */
   /* =============================================================================================== */

   parse arg str , thisnum , chrsW , chrsQ
   if chrsW == '' then chrsW = ' '||d2c(9)
   if chrsQ == '' then chrsQ = '"'||"'"||'`'
   i         = 1
   quote     = ''
   numwords  = 0
   Thisword  = ''
   w1        = 0
   w2        = 0

   /* find start of first word by trimming blanks and word characters                                 */
   do while i <= length(str) ,
          & pos(substr(str, i, 1), chrsW) > 0
      i = i +1
   end
   if i <= length(str) then numwords = numwords +1
   if numwords == thisnum then w1 = i

   do while (i <= length(str))
      if pos(substr(str, i, 1), chrsQ) \= 0 then do
            /* start of quoted string;
             nothing counts until we get to the matching quote (or end of string) */
            quote = substr(str, i, 1)
            tsti = i +1
            /*  bugfix for "-opt=''"                                                                  */
            do while tsti < length(str) ,
                   & substr(str, tsti, 1) \= quote
               tsti = tsti +1
            end
            if substr(str, tsti, 1) == quote then
               /* found the matching quote                                                            */
               i = tsti
            else
               /* ran out of string with no closing quote, so ignore this quote                       */
               nop
            i = i +1
         end   /*  quote  */

      if pos(substr(str, i, 1), chrsW) \= 0 ,
       & i <= length(str) then do
            /* a word separator, skip all blanks, find start of next word                             */
            /* unless the end of the word we want, then quit                                          */
            if w1 > 0 then do
                  w2 = i -1
                  leave
               end
            do while (pos(substr(str, i, 1), chrsW) \= 0) ,
                   & (i <= length(str))
               i = i +1
            end
            if i <= length(str) then numwords = numwords +1
            /* if this is the start of the word we want, note position in string                      */
            if numwords == thisnum then w1 = i
         end   /*  wordsep  */
      else
         /* a normal character                                                                        */
         i = i +1
   end   /*  do i  */
   if i > length(str) then i = length(str)
   if w1 > 0 then do
         /* we know w1, if we ran out of string, w2 is the end of string                              */
         if w2 == 0 then w2 = i
         /* found start and end of word, copy it                                                      */
         wlen = w2 - w1 +1
         Thisword = substr(str, w1, wlen)
      end
   return Thisword
   /* === End of WordQ ============================================================================== */

   WordsQ: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     foo = WordsQ(string, <separator characters>, <quote characters>)                            */
   /*                                                                                                 */
   /*  The last two parameters are optional, so it will work                                          */
   /*  as a drop-in replacement for the Rexx Word() function.                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  A better Words() function--                                                                    */
   /*  It recognizes a quoted string as one word                                                      */
   /*  Optional parameters let you decide which characters                                            */
   /*  count as word separators and as quotes.                                                        */
   /*                                                                                                 */
   /*  Default word separators are <space> and <tab>                                                  */
   /*  Default quote characters are single, double, and                                               */
   /*  back quotes-- ' " `                                                                            */
   /* =============================================================================================== */

   parse arg str , chrsW , chrsQ
   if chrsW == '' then chrsW = ' '||d2c(9)
   if chrsQ == '' then chrsQ = '"'||"'"||'`'
   i         = 1
   quote     = ''
   numwords  = 0

   /* find start of first word by trimming blanks and word characters                                 */
   do while i <= length(str) ,
          & pos(substr(str, i, 1), chrsW) > 0
      i = i +1
   end
   if i <= length(str) then numwords = numwords +1

   do while (i <= length(str))
      if pos(substr(str, i, 1), chrsQ) \= 0 then do
            /* start of quoted string;
             nothing counts until we get to the matching quote (or end of string) */
            quote = substr(str, i, 1)
            tsti = i +1
            /*  bugfix for "-opt=''"                                                                  */
            do while tsti < length(str) ,
                   & substr(str, tsti, 1) \= quote
               tsti = tsti +1
            end
            if substr(str, tsti, 1) == quote then
               /* found the matching quote                                                            */
               i = tsti
            else
               /* ran out of string with no closing quote, so ignore this quote                       */
               nop
            i = i +1
         end   /*  quote  */

      if pos(substr(str, i, 1), chrsW) \= 0 ,
       & i <= length(str) then do
            /* a word separator, skip all blanks, find start of next word                             */
            do while (pos(substr(str, i, 1), chrsW) \= 0) ,
                   & (i <= length(str))
               i = i +1
            end
            if i <= length(str) then numwords = numwords +1
         end
      else
         /* a normal character                                                                        */
         i = i +1
   end   /*  do i  */
   return numwords
   /* === End of WordsQ ============================================================================= */


